/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 * Copyright (c) 1991 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Graphics server event routines for windows.
 */
#include "windows.h"
#include "wintern.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Update mouse status and issue events on it if necessary.
 * This function doesn't block, but is normally only called when
 * there is known to be some data waiting to be read from the mouse.
 */
void GsCheckMouseEvent(void)
{
	COORD		rootx;		/* latest mouse x position */
	COORD		rooty;		/* latest mouse y position */
	BUTTON		newbuttons;	/* latest buttons */
	int		mousestatus;	/* latest mouse status */

	/* Read the latest mouse status: */
	mousestatus = GdReadMouse(&rootx, &rooty, &newbuttons);
	if(mousestatus < 0) {
		/*GsError(GR_ERROR_MOUSE_ERROR, 0);*/
		return;
	} else if(mousestatus) /* Deliver events as appropriate: */	
		GsHandleMouseStatus(rootx, rooty, newbuttons);
}

/*
 * Update keyboard status and issue events on it if necessary.
 * This function doesn't block, but is normally only called when
 * there is known to be some data waiting to be read from the keyboard.
 */
void GsCheckKeyboardEvent(void)
{
	unsigned char	ch;		/* latest character */
	MODIFIER	modifiers;	/* latest modifiers */
	int		keystatus;	/* latest keyboard status */

	/* Read the latest keyboard status: */
	keystatus = GdReadKeyboard(&ch, &modifiers);
	if(keystatus < 0) {
		if(keystatus == -2)	/* special case for ESC pressed*/
			GsTerminate();
		/*GsError(GR_ERROR_KEYBOARD_ERROR, 0);*/
		return;
	} else if(keystatus) /* Deliver events as appropriate: */	
		GsDeliverKeyboardEvent(ch, modifiers);
}

/*
 * Handle all mouse events.  These are mouse enter, mouse exit, mouse
 * motion, mouse position, button down, and button up.  This also moves
 * the cursor to the new mouse position and changes it shape if needed.
 */
void GsHandleMouseStatus(COORD newx, COORD newy, BUTTON newbuttons)
{
	BUTTON		changebuttons;	/* buttons that have changed */
	unsigned char	ch;		/* latest character */
	MODIFIER	modifiers;	/* latest modifiers */
	static BUTTON curbuttons;

	GdReadKeyboard(&ch, &modifiers); /* Read the modifiers */

	/*
	 * First, if the mouse has moved, then position the cursor to the
	 * new location, which will send mouse enter, mouse exit, focus in,
	 * and focus out events if needed.  Check here for mouse motion and
	 * mouse position events.
	 */
	if (newx != cursorx || newy != cursory) {
		GsMoveCursor(newx, newy);
		GsDeliverMotionEvent(newbuttons, modifiers);
	}

	/*
	 * Next, generate a button up event if any buttons have been released.
	 */
	changebuttons = (curbuttons & ~newbuttons);
	if (changebuttons)
		GsDeliverButtonEvent(newbuttons, changebuttons, modifiers);

	/*
	 * Finally, generate a button down event if any buttons have been
	 * pressed.
	 */
	changebuttons = (~curbuttons & newbuttons);
	if (changebuttons)
		GsDeliverButtonEvent(newbuttons, changebuttons, modifiers);

	curbuttons = newbuttons;
}

/*
 * Deliver a mouse button event.
 */
void 
GsDeliverButtonEvent(BUTTON buttons, BUTTON changebuttons, MODIFIER modifiers)
{
	POINT	pt;
	int	hittest;

	pt.x = cursorx;
	pt.y = cursory;
	ScreenToClient(mousewp, &pt);
	hittest = SendMessage(mousewp, WM_NCHITTEST, 0,
		MAKELONG(cursorx, cursory));
	if(changebuttons & LBUTTON) {
		PostMessage(mousewp,
			(buttons&LBUTTON)? WM_NCLBUTTONDOWN: WM_NCLBUTTONUP,
			hittest, MAKELONG(cursorx, cursory));
		if(hittest == HTCLIENT)
		  PostMessage(mousewp,
			(buttons&LBUTTON)? WM_LBUTTONDOWN: WM_LBUTTONUP, 0,
			MAKELONG(pt.x, pt.y));
	}

	if(changebuttons & RBUTTON) {
		PostMessage(mousewp,
			(buttons&RBUTTON)? WM_NCRBUTTONDOWN: WM_NCRBUTTONUP,
			hittest, MAKELONG(cursorx, cursory));
		if(hittest == HTCLIENT)
		  PostMessage(mousewp,
			(buttons&RBUTTON)? WM_RBUTTONDOWN: WM_RBUTTONUP, 0,
			MAKELONG(pt.x, pt.y));
	}
}

/*
 * Deliver a mouse motion event.
 */
void GsDeliverMotionEvent(BUTTON buttons, MODIFIER modifiers)
{
	POINT	pt;
	int	hittest;

	hittest = SendMessage(mousewp, WM_NCHITTEST, 0,
		MAKELONG(cursorx, cursory));
	PostMessage(mousewp, WM_NCMOUSEMOVE, 0, MAKELONG(cursorx, cursory));
	if(hittest == HTCLIENT) {
		pt.x = cursorx;
		pt.y = cursory;
		ScreenToClient(mousewp, &pt);
		PostMessage(mousewp, WM_MOUSEMOVE, 0, MAKELONG(pt.x, pt.y));
	}
}

/*
 * Deliver a keyboard event.
 */
void GsDeliverKeyboardEvent(UCHAR ch, MODIFIER modifiers)
{
	SendMessage(mousewp, WM_CHAR, (WPARAM)ch, 0L);
}

/*
 * Try to deliver a exposure event to the clients which have selected for it.
 * This does not send exposure events for unmapped windows.
 * Exposure events do not propagate upwards.
 */
void
GsDeliverExposureEvent(HWND wp, COORD x, COORD y, COORD width, COORD height)
{
	if (wp->unmapcount)
		return;
	PostMessage(wp, WM_PAINT, 0, 0L);
}

/*
 * Move the cursor to the specified absolute screen coordinates.
 * The coordinates are that of the defined hot spot of the cursor.
 * The cursor's appearance is changed to that defined for the window
 * in which the cursor is moved to.  In addition, mouse enter, mouse
 * exit, focus in, and focus out events are generated if necessary.
 */
void GsMoveCursor(COORD x, COORD y)
{
	/*
	 * Move the cursor only if necessary, offsetting it to
	 * place the hot spot at the specified coordinates.
	 */
	if (x != cursorx || y != cursory) {
		if(curcursor)
			GdMoveCursor(x - curcursor->cursor.hotx,
				y - curcursor->cursor.hoty);
		cursorx = x;
		cursory = y;
	}

	/*
	 * Now check to see which window the mouse is in and whether or
	 * not the cursor shape should be changed.
	 */
	GsCheckMouseWindow();
	GsCheckCursor();
}

/*
 * Check to see if the cursor shape is the correct shape for its current
 * location.  If not, its shape is changed.
 */
void GsCheckCursor(void)
{
	HWND 		wp;		/* window cursor is in */
	HCURSOR		cp;		/* cursor definition */

	/*
	 * Get the cursor at its current position, and if it is not the
	 * currently defined one, then set the new cursor.  However,
	 * if the window is currently captured, then leave it alone.
	 */
	wp = capturewp;
	if (wp == NULL)
		wp = mousewp;

	cp = wp->cursor;
	if (cp == curcursor)
		return;

	/*
	 * It needs redefining, so do it.
	 */
	curcursor = cp;
	GdMoveCursor(cursorx - cp->cursor.hotx, cursory - cp->cursor.hoty);
	GdSetCursor(&cp->cursor);
}

/*
 * Find the window which is currently visible for the specified coordinates.
 * This just walks down the window tree looking for the deepest mapped
 * window which contains the specified point.  If the coordinates are
 * off the screen, the root window is returned.
 */
HWND
WndFindVisibleWindow(COORD x, COORD y)
{
	HWND	wp;		/* current window */
	HWND	retwp;		/* returned window */

	wp = rootwp;
	retwp = wp;
	while (wp) {
		if (!wp->unmapcount &&
		    wp->winrect.left <= x && wp->winrect.top <= y &&
		    wp->winrect.right > x && wp->winrect.bottom > y) {
			retwp = wp;
			wp = wp->children;
			continue;
		}
		wp = wp->siblings;
	}
	return retwp;
}

/*
 * Check to see if the window the mouse is currently in has changed.
 * If so, generate enter and leave events as required.  The newest
 * mouse window is remembered in mousewp.  However, do not change the
 * window if it is captured.
 */
void GsCheckMouseWindow(void)
{
	HWND	wp;		/* newest window for mouse */

	wp = dragwp;
	if(!wp)
		wp = capturewp;
	if(!wp)
		wp = WndFindVisibleWindow(cursorx, cursory);
	mousewp = wp;
}

/*
 *  Copy dstsiz bytes, including nul, from src to dst.
 *  Return # bytes, excluding nul, copied.
 */
int
strzcpy(char *dst,const char *src,int dstsiz)
{
	int	cc = dstsiz;

	/* return 0 on NULL src*/
	if(!src)
		cc = dstsiz = 1;

	while(--dstsiz > 0) {
		if((*dst++ = *src++) == '\0')
			return cc - dstsiz - 1;
	}
	*dst = 0;
	return cc - dstsiz - 1;
}
